structure Dynamics =
struct
  structure L = LambdaPlusPlus
  exception TypeError of string

  (*  subst : (int * L.exp) * L.exp -> L.exp
   *  subst ((x, e), ee) = substitutes all instances of the variable x
   *    in ee with the value e
   *  not curried so it can be used with foldl                              *)
  fun subst (x, e) ee =
    case ee of
      L.Variable s => if s = x then e else ee
    | L.Lambda (s, e1) => if s <> x then L.Lambda (s, subst (x, e) e1) else ee
    | L.Apply (e1, e2) => L.Apply (subst (x, e) e1, subst (x, e) e2)
    | L.Tuple (e1, e2) => L.Tuple (subst (x, e) e1, subst (x, e) e2)
    | L.First e1 => L.First (subst (x, e) e1)
    | L.Second e1 => L.Second (subst (x, e) e1)
    | L.Left e1 => L.Left (subst (x, e) e1)
    | L.Right e1 => L.Right (subst (x, e) e1)
    | L.Case (e', (s1, e1), (s2, e2)) =>
        L.Case (
          subst (x, e) e',
          (s1, if s1 <> x then subst (x, e) e1 else e1),
          (s2, if s2 <> x then subst (x, e) e2 else e2)
        )
  fun subst' (xe, ee) = subst xe ee (* uncurried version *)

  (*  eval : L.exp -> L.exp
   *  eval e = the evaluation result of e (must be a value) *)
  fun eval _ = raise Fail "unimplemented"

end
